-- This Source Code Form is subject to the terms of the bCDDL, v. 1.1.
-- If a copy of the bCDDL was not distributed with this
-- file, You can obtain one at http://beamng.com/bCDDL-1.1.txt

local im  = ui_imgui

local C = {}

C.name = 'Update extra AI race mission'
C.description = 'Update extra AI race mission rival racer state.'
C.category = 'once_f_duration'
C.color = im.ImVec4(0.4, 0.9, 1.0, 0.9)
C.author = 'mathkuro'
C.tmpSecondPassFlag = true
C.pinSchema = {
  { dir = 'in', type = 'table', name = 'raceData', tableType = 'raceData', description = 'Data from the race for other nodes to process.' },
  { dir = 'in', type = 'table', name = 'navgraphPath', tableType = 'navgraphPath', description = 'Navgraph path for the AI to follow.' },
  {dir = 'in', type = 'number', name = 'maxLap', description = 'Max lap.'},
  {dir = 'in', type = 'number', name = 'racerAggression', description = 'Racer aggression.'},
  {dir = 'in', type = 'number', name = 'racerSkill', description = 'Racer skill.'},
  { dir = 'in', type = 'bool', name = 'avoidCars', hidden = true, default = true, description = 'Whether or not the vehicle should try to avoid collisions.' },
  {dir = 'in', type = 'bool', name = 'useRubberBand', description = 'Use rubber band mode.'},
  {dir = 'in', type = 'string', name = 'driveStyle', description = 'Driving style (default, offRoad, etc.).'},
  {dir = 'in', type = 'number', name = 'randomnessScale', description = 'randomness scale of AI parameters.'},
  {dir = 'in', type = 'bool', name = 'enableAIRespawn', description = 'Enable AI respawn.'},

  { dir = 'out', type = 'string', name = 'finalWp', hidden = true, description = 'Name of the final waypoint in the route.' }
}

C.tags = {}

function C:init()
  self.speedThreshold = 3
  self.damageThreshold = 500
  self.stopFrameThreshold = 600
  self.aiRestartDelayThreshold = 60
  self.defaultAIPath = {}

  self:onNodeReset()
end

function C:onNodeReset()
  self.state = 'inactive'
end

function C:startAIRacer()
  self.lapCount = math.max(1, self.pinIn.maxLap.value or 1)

  if self.pinIn.navgraphPath.value then
    self.aiPath = self.pinIn.navgraphPath.value.aiPath or self.pinIn.navgraphPath.value -- backwards compatibility
  end

  if not self.aiPath[1] then
    return
  end

  self.pinOut.finalWp.value = self.aiPath[#self.aiPath]
  local aiPath = shallowcopy(self.aiPath)
  if self.lapCount > 1 then
    table.insert(aiPath, aiPath[1]) -- allows AI to continue to the next lap
  end

  self.defaultAIPath = aiPath

  local wpSpeeds = {}
  if not self.pinIn.stopAtFinish.value then
    wpSpeeds[self.pinOut.finalWp.value] = 100 -- last waypoint
  end

  local racerInfoTable = mathkuro_extraRaceMission.getRacerInfoTable()
  for vid, _ in pairs(racerInfoTable) do
    local _veh = be:getObjectByID(vid)
    if _veh then
      local cmd = 'ai.driveUsingPath({wpTargetList = ' .. serialize(aiPath) .. ', wpSpeeds = ' .. serialize(wpSpeeds) .. ', noOfLaps = ' .. self.lapCount .. '})'
      _veh:queueLuaCommand(cmd)
      print(cmd)

      mathkuro_extraRaceMission.updateRacerInfo(vid, nil, nil, nil, aiPath)

      self:setAIParameters(vid)

      _veh:queueLuaCommand('controller.setFreeze(0)')

      log('D', 'updateExtraAIRaceNode', tostring(vid) .. ' is updated.')
    end
  end

  self:updateCheckPointData(vid)

  self.state = 'started'
  log('I', 'updateExtraAIRaceNode', 'AI Racer is started.')
end

function C:work()
  if self.state == 'inactive' then
    self:startAIRacer()
  elseif self.pinIn.enableAIRespawn.value then
    local maxLap = self.pinIn.maxLap.value or 1

    local racerInfoTable = mathkuro_extraRaceMission.getRacerInfoTable()

    for vid, racerInfo in pairs(racerInfoTable) do
      local veh = be:getObjectByID(vid)
      if racerInfo.curLap and racerInfo.curLap >= maxLap then
        -- if the racer route completed
        if racerInfo.curLap == maxLap then
          veh:queueLuaCommand('ai.setMode("random")')
          racerInfo.curLap = racerInfo.curLap + 1
        end
      else
        if self:updateCheckPointData(vid) then
          veh:queueLuaCommand('recovery.homePoint = {pos=vec3(obj:getPositionXYZ()), dirFront=obj:getDirectionVector(), dirUp=obj:getDirectionVectorUp()}')
        end

        self:updateAIParameters(vid, racerInfo)

        self:restartAIRacer(vid, racerInfo)
      end
    end
  end
end

function C:updateCheckPointData(vid)
  local curCP = nil
  local curLap = nil
  local changed = false
  -- ref: lua\ge\extensions\flowgraph\nodes\gameplay\race\raceCpLapChange.lua
  local state = self.pinIn.raceData.value.states[vid]
  if not state then return end
  local events = state.events
  if not events then return end

  if events.rollingStarted or events.pathnodeReached or events.raceStarted or events.lapComplete or events.raceComplete then
    local race = self.pinIn.raceData.value
    local curSeg1 = state.currentSegments[1]
    if curSeg1 then
      local graphSeg = race.path.config.graph[curSeg1]
      curCP = graphSeg.linearCPIndex or nil
    else
      curCP = nil
    end

    curLap = state.currentLap
    log('D', 'updateExtraAIRaceNode', 'curLap: ' .. tostring(curLap))
    changed = true

    mathkuro_extraRaceMission.updateRacerInfo(vid, curCP, curLap)
  end

  return changed
end

function C:restartAIRacer(vid, racerInfo)
  local veh = be:getObjectByID(vid)

  local daylightValues = {0.22, 0.78} -- sunset & sunrise

  local oldAIPath = racerInfo.curAIPath or self.defaultAIPath
  local aiPath = {}
  -- game speed( before unit convertion)
  local _vehData = map.objects[vid]
  if _vehData == nil then
    return
  end

  local _curSpeed = _vehData.vel:length()
  local _curDamage = _vehData.damage
  -- print('_curSpeed: ' .. tostring(_curSpeed) .. ', _curDamage: ' .. tostring(_curDamage))

  if racerInfo.isRecovered then
    if not racerInfo.aiRestartDelayCount then
      racerInfo.aiRestartDelayCount = 0
    end
    racerInfo.aiRestartDelayCount = racerInfo.aiRestartDelayCount + 1

    if self.aiRestartDelayThreshold < racerInfo.aiRestartDelayCount then
      racerInfo.aiRestartDelayCount = 0
      racerInfo.isRecovered = false
      racerInfo.stopFrameCount = 0

      veh:queueLuaCommand('electrics.set_lightbar_signal(1)')

      local currentTODTime = core_environment.getTimeOfDay()['time']
      print(currentTODTime)
      if daylightValues[1] < currentTODTime and currentTODTime < daylightValues[2] then
        veh:queueLuaCommand('electrics.setLightsState(2)')
        print('electrics.setLightsState(2)')
      end

      if not racerInfo.curCP then
        aiPath = oldAIPath
      else
        print('curLap: ' .. tostring(racerInfo.curLap) .. ', maxLap: ' .. tostring(racerInfo.maxLap) .. ', curCP: ' .. tostring(racerInfo.curCP) .. ', maxCP: ' .. tostring(racerInfo.maxCP))
        print('oldAIPath: ' .. serialize(oldAIPath))
        for k, v in pairs(oldAIPath) do
          if k >= racerInfo.curCP then
            table.insert(aiPath, v)
          end
        end
        for k, v in pairs(oldAIPath) do
          if k < racerInfo.curCP then
            table.insert(aiPath, v)
          end
        end
        if self.lapCount >= 1 then
          -- If the path is closed
          table.insert(aiPath, oldAIPath[racerInfo.curCP])
        end
      end

      local cmd = "ai.driveUsingPath({wpTargetList = " .. serialize(aiPath) .. ", noOfLaps = " .. tostring(self.lapCount) .. ", aggression = " .. tostring(self.pinIn.racerAggression.value)  .. ", avoidCars = 'on'})"
      veh:queueLuaCommand(cmd)
      print(cmd)

      mathkuro_extraRaceMission.updateRacerInfo(vid, nil, nil, nil, aiPath)

      self:setAIParameters(vid)
      return
    end

    return
  end

  -- If the racer does not reach first checkpoint.
  if not racerInfo.curLap or (racerInfo.curLap >= self.pinIn.maxLap.value and not racerInfo.curCP) then
    return
  end

  -- If the speed is below threshold reset certain speed 
  if not racerInfo.stopFrameCount or self.speedThreshold < _curSpeed then
    racerInfo.stopFrameCount = 0
    return
  end
  racerInfo.stopFrameCount = racerInfo.stopFrameCount + 1

  -- If the rival vehicle stop long time, recover it to last check point
  if racerInfo.stopFrameCount < self.stopFrameThreshold then
    return
  end
  print('Rival Racer: ' .. tostring(vid) .. ' is recovered.')
  racerInfo.isRecovered = true
  veh:queueLuaCommand('recovery.loadHome()')
end

function C:setAIParameters(vid)
  -- ref: lua\ge\extensions\flowgraph\nodes\vehicle\ai\aiParams.lua

  local veh = be:getObjectByID(vid)
  if not veh then return end

  local randomnessScale = self.pinIn.randomnessScale.value or 0.0

  if self.pinIn.racerAggression.value ~= nil then
    local racerAggression = self.pinIn.racerAggression.value
    racerAggression = math.random(
      math.max(0.3, racerAggression - 0.03 * randomnessScale),
      math.min(1.5, racerAggression + 0.03 * randomnessScale)
    )
    veh:queueLuaCommand('ai.setAggression('..clamp(racerAggression,0.1,2.0)..')')

    mathkuro_extraRaceMission.updateRacerInfo(vid, nil, nil, racerAggression)
  end

  if self.pinIn.avoidCars.value ~= nil then
    veh:queueLuaCommand('ai.setAvoidCars("'..(self.pinIn.avoidCars.value and 'on' or 'off')..'")')
  end

  if self.pinIn.useRubberBand.value ~= nil then
    veh:queueLuaCommand('ai.setAggressionMode("'..(self.pinIn.useRubberBand.value and 'rubberBand' or 'off')..'")')
  end

  if self.pinIn.racerSkill.value then
    local racerSkill = self.pinIn.racerSkill.value
    local edgeDist = math.random(
      math.max(0, 0.6 * (1 - racerSkill) - 0.1 * randomnessScale),
      math.min(1, 0.6 * (1 - racerSkill) + 0.1 * randomnessScale)
    )
    veh:queueLuaCommand('ai.setParameters({edgeDist = '..edgeDist..'})')

    local awarenessForce = math.random(
      math.max(0.01, 0.36 - (0.3 * racerSkill) - 0.03 * randomnessScale),
      math.min(0.5, 0.36 - (0.3 * racerSkill) + 0.03 * randomnessScale)
    )
    veh:queueLuaCommand('ai.setParameters({awarenessForceCoef = '..awarenessForce..'})')

    local turnForce = math.random(
      math.max(0.01, 4 * (racerSkill * racerSkill) - 1.5 * randomnessScale),
      math.min(10, 4 * (racerSkill * racerSkill) + 1.5 * randomnessScale)
    )
    veh:queueLuaCommand('ai.setParameters({turnForceCoef = '..turnForce..'})')
  end

  if self.pinIn.driveStyle.value then
    veh:queueLuaCommand('ai.setParameters({driveStyle = "'..self.pinIn.driveStyle.value..'"})')
  end
end

function C:updateAIParameters(vid, racerInfo)
  -- ref: lua\ge\extensions\flowgraph\nodes\gameplay\race\raceNodeReached.lua

  local veh = be:getObjectByID(vid)
  if not veh then return end

  local race = self.pinIn.raceData.value
  local events = race.states[vid].events
  if not events then return end

  if events.pathnodeReached then
    local pn = race.path.pathnodes.objects[events.pathnodeReachedId]
    local aiAggression, valType, succ = pn.customFields:get('aiAggression')
    if succ and aiAggression > 0 then
      aiAggression = min(aiAggression, racerInfo.aiAggression)
    else
      aiAggression = racerInfo.aiAggression
    end

    if aiAggression then
      veh:queueLuaCommand('ai.setAggression('..clamp(aiAggression,0.1,2.0)..')')
    else
      log('W', 'updateExtraAIRaceNode', 'aiAggression is nil')
    end
  end
end

return _flowgraph_createNode(C)
